/* SD Card Raw Read Demo
 * http://frank.circleofcurrent.com/musicclock/
 * Copyright (c) 2011 Frank Zhao
 *
 * this example is based off sd_raw from http://www.roland-riegel.de/sd-reader/index.html
 * 
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of either the GNU General Public License version 2
 * or the GNU Lesser General Public License version 2.1, both as
 * published by the Free Software Foundation.
 *
 */

// please see http://www.nongnu.org/avr-libc/user-manual/modules.html for info on these modules

#include <avr/io.h> // allows access to AVR built-in peripherals and registers
#include <avr/pgmspace.h> // allows access to flash memory
#include <util/delay.h> // provides delays
#include <stdio.h> // allows formatted debug output
#include <avr/power.h> // used for setting clock prescaler
#include "sd_raw_defs.h" // some definitions
#include "usb_serial/usb_serial.h" // provides USB-to-serial functionality

/* enable outputs for MOSI, SCK, SS, input for MISO */
// configure pins here
#define configure_pin_mosi()    DDRB |= _BV(2)
#define configure_pin_sck()     DDRB |= _BV(1)
#define configure_pin_ss()      DDRB |= _BV(0)
#define configure_pin_miso()    DDRB &= ~_BV(3)
#define select_card()           PORTB &= ~_BV(0)
#define unselect_card()         PORTB |= _BV(0)

// this global stores the current card type
static uint8_t sd_raw_card_type;

// this is a wrapper for usb_putchar that allows me to use printf from stdio.h
static inline int debug_putchar(char c, FILE* stream)
{
    return usb_serial_putchar(c); // refer to http://www.pjrc.com/teensy/usb_serial.html
}

// setup stream for printf from stdio.h
static FILE mystdout = FDEV_SETUP_STREAM(debug_putchar, NULL, _FDEV_SETUP_WRITE);

// initialize built-in SPI peripheral
void spi_init()
{
    /* enable outputs for MOSI, SCK, SS, input for MISO */
    configure_pin_mosi();
    configure_pin_sck();
    configure_pin_ss();
    configure_pin_miso();
    
    /* initialize SPI with lowest frequency; max. 400kHz during identification mode of card */
    SPCR = (0 << SPIE) | /* SPI Interrupt Disable */
           (1 << SPE)  | /* SPI Enable */
           (0 << DORD) | /* Data Order: MSB first */
           (1 << MSTR) | /* Master mode */
           (0 << CPOL) | /* Clock Polarity: SCK low when idle */
           (0 << CPHA) | /* Clock Phase: sample on rising SCK edge */
           (1 << SPR1) | /* Clock Frequency: f_OSC / 128 */
           (1 << SPR0);
    SPSR &= ~(1 << SPI2X); /* No doubled clock frequency */
}

// set SPI clock for low frequency
inline void spi_lowfreq()
{
    SPCR = (0 << SPIE) | /* SPI Interrupt Disable */
           (1 << SPE)  | /* SPI Enable */
           (0 << DORD) | /* Data Order: MSB first */
           (1 << MSTR) | /* Master mode */
           (0 << CPOL) | /* Clock Polarity: SCK low when idle */
           (0 << CPHA) | /* Clock Phase: sample on rising SCK edge */
           (1 << SPR1) | /* Clock Frequency: f_OSC / 128 */
           (1 << SPR0);
    SPSR &= ~(1 << SPI2X); /* No doubled clock frequency */
}

// set SPI clock for high frequency
inline void spi_highfreq()
{
    SPCR = (0 << SPIE) | /* SPI Interrupt Disable */
           (1 << SPE)  | /* SPI Enable */
           (0 << DORD) | /* Data Order: MSB first */
           (1 << MSTR) | /* Master mode */
           (0 << CPOL) | /* Clock Polarity: SCK low when idle */
           (0 << CPHA) | /* Clock Phase: sample on rising SCK edge */
           (0 << SPR1) | /* Clock Frequency: f_OSC / 4 */
           (0 << SPR0);
    SPSR |= (1 << SPI2X); /* Doubled clock frequency, f_osc / 2 */
}

// read a byte through SPI
inline uint8_t spi_rx()
{
    SPDR = 0xFF; // start blank transfer
    loop_until_bit_is_set(SPSR, SPIF); // wait until transfer finishes
    return SPDR; // return result
}

// send a byte through SPI
inline uint8_t spi_tx(uint8_t d)
{
    SPDR = d; // send
    loop_until_bit_is_set(SPSR, SPIF); // wait until transfer finishes
    return SPDR;
}

// utility to send commands to SD card
/**
 * \ingroup sd_raw
 * Send a command to the memory card which responses with a R1 response (and possibly others).
 *
 * \param[in] command The command to send.
 * \param[in] arg The argument for command.
 * \returns The command answer.
 */
uint8_t sd_raw_send_command(uint8_t command, uint32_t arg)
{
    uint8_t response;

    /* wait some clock cycles */
    spi_rx();

    /* send command via SPI */
    spi_tx(0x40 | command);
    spi_tx((arg >> 24) & 0xff);
    spi_tx((arg >> 16) & 0xff);
    spi_tx((arg >> 8) & 0xff);
    spi_tx((arg >> 0) & 0xff);
    switch(command)
    {
        case CMD_GO_IDLE_STATE:
           spi_tx(0x95);
           break;
        case CMD_SEND_IF_COND:
           spi_tx(0x87);
           break;
        default:
           spi_tx(0xff);
           break;
    }
    
    /* receive response */
    for(uint8_t i = 0; i < 10; ++i)
    {
        response = spi_rx();
        if(response != 0xff)
            break;
    }

    return response;
}

// this function reads blocks of data from the SD card
/**
 * \ingroup sd_raw
 * Reads raw data from the card.
 *
 * \param[in] offset The offset from which to read.
 * \param[out] buffer The buffer into which to write the data.
 * \param[in] length The number of bytes to read.
 * \returns 0 on failure, 1 on success.
 * \see sd_raw_read_interval, sd_raw_write, sd_raw_write_interval
 */
uint8_t sd_raw_read(offset_t offset, uint8_t* buffer, uintptr_t length)
{
    offset_t block_address;
    uint16_t block_offset;
    uint16_t read_length;
    while(length > 0)
    {
        /* determine byte count to read at once */
        block_offset = offset & 0x01ff;
        block_address = offset - block_offset;
        read_length = 512 - block_offset; /* read up to block border */
        if(read_length > length)
            read_length = length;
        
        /* address card */
        select_card();

        /* send single block request */

        if(sd_raw_send_command(CMD_READ_SINGLE_BLOCK, (sd_raw_card_type & (1 << SD_RAW_SPEC_SDHC) ? block_address / 512 : block_address)))
        {
            unselect_card();
            return 0;
        }

        /* wait for data block (start byte 0xfe) */
        while(spi_rx() != 0xfe);

        /* read byte block */
        uint16_t read_to = block_offset + read_length;
        for(uint16_t i = 0; i < 512; ++i)
        {
            uint8_t b = spi_rx();
            if(i >= block_offset && i < read_to)
                *buffer++ = b;
        }
        
        /* read crc16 */
        spi_rx();
        spi_rx();
        
        /* deaddress card */
        unselect_card();

        /* let card some time to finish */
        spi_rx();

        length -= read_length;
        offset += read_length;
    }

    return 1;
}

// program entry point
int main()
{
    clock_prescale_set(clock_div_2); // this is because we are running at 3.3V
    // see http://www.nongnu.org/avr-libc/user-manual/group__avr__power.html and http://www.pjrc.com/teensy/3volt.html
    
    stdout = &mystdout; // setup stream for printf from stdio.h
    
    // initialize stuff
    usb_init();
    
    while (usb_configured() == 0) _delay_ms(5); // waits until USB is ready
    
    _delay_ms(5000); // 5 seconds for you to connect hyperterminal
    
    // here's the start of the demonstration
    printf("SD Card Raw Read Demo\r\n");
    
    spi_init();
    
    unselect_card();
    
    printf("Resetting Card\r\n");
    
    /* card needs 74 cycles minimum to start up */
    for(char i = 0; i < 10; i++)
    {
        spi_rx();
    }
    
    /* address card */
    select_card();

    /* reset card */
    uint8_t response;    
    for(uint16_t i = 0; ; ++i)
    {
        response = sd_raw_send_command(CMD_GO_IDLE_STATE, 0);
        
        if(response == (1 << R1_IDLE_STATE))
            break;

        if(i == 0x1ff)
        {
            unselect_card();
            printf("Error: SD card won't go idle\r\n");
            while (1); // freeze due to error
        }
    }
    printf("Response: %d\r\n", response);
   
    /* determine SD/MMC card type */
    printf("Reading Card Type\r\n");
    response = sd_raw_send_command(CMD_SEND_IF_COND, 0x100 /* 2.7V - 3.6V */ | 0xaa /* test pattern */);
    if((response & (1 << R1_ILL_COMMAND)) == 0)
    {
        spi_rx();
        spi_rx();
        
        if((spi_rx() & 0x01) == 0)
        {
            /* card operation voltage range doesn't match */
            printf("Error: card operation voltage range doesn't match\r\n");
            while (1);
            return 0; 
        }
        
        if(spi_rx() != 0xaa)
        {
            /* wrong test pattern */
            printf("Error: Wrong test pattern\r\n");
            while (1); // freeze due to error
        }

        /* card conforms to SD 2 card specification */
        sd_raw_card_type |= (1 << SD_RAW_SPEC_2);
        printf("Card Type: SD Spec 2\r\n");
    }
    else
    {
        sd_raw_send_command(CMD_APP, 0);
        response = sd_raw_send_command(CMD_SD_SEND_OP_COND, 0);
        if((response & (1 << R1_ILL_COMMAND)) == 0)
        {
            /* card conforms to SD 1 card specification */
            sd_raw_card_type |= (1 << SD_RAW_SPEC_1);
            printf("Card Type: SD Spec 1\r\n");
        }
        else
        {
            printf("Card Type: MMC\r\n");
        }
    }
    
    /* wait for card to get ready */
    printf("Waiting for card to be ready\r\n");
    for(uint16_t i = 0; ; ++i)
    {
        if(sd_raw_card_type & ((1 << SD_RAW_SPEC_1) | (1 << SD_RAW_SPEC_2)))
        {
            uint32_t arg = 0;
            if(sd_raw_card_type & (1 << SD_RAW_SPEC_2))
                arg = 0x40000000;
            sd_raw_send_command(CMD_APP, 0);
            response = sd_raw_send_command(CMD_SD_SEND_OP_COND, arg);
        }
        else
        {
            response = sd_raw_send_command(CMD_SEND_OP_COND, 0);
        }

        if((response & (1 << R1_IDLE_STATE)) == 0)
            break;

        if(i == 0x7fff)
        {
            unselect_card();
            return 0;
        }
    }

    if(sd_raw_card_type & (1 << SD_RAW_SPEC_2))
    {
        if(sd_raw_send_command(CMD_READ_OCR, 0))
        {
            unselect_card();
            return 0;
        }

        if(spi_rx() & 0x40)
            sd_raw_card_type |= (1 << SD_RAW_SPEC_SDHC);

        spi_rx();
        spi_rx();
        spi_rx();
    }

    /* set block size to 512 bytes */
    printf("Setting block size = 512\r\n");
    if(sd_raw_send_command(CMD_SET_BLOCKLEN, 512))
    {
        unselect_card();
        printf("Error: failed to set block size\r\n");
        while (1); // freeze for error
    }

    /* deaddress card */
    unselect_card();

    /* switch to highest SPI frequency possible */
    spi_highfreq();

    uint8_t testdata[512];
    
    printf("Begin Raw Card Read (3 blocks):\r\n");
    
    for (int i = 0; i < 3; i++)
    {
        printf("Block[%d]:\r\n", i);
        if(!sd_raw_read(i * 512, testdata, 512))
        {
            printf("Error: block read failed\r\n");
            while (1); // freeze for error
        }
        
        for (int j = 0; j < 512; j++)
        {
            debug_putchar(testdata[j], 0);
        }
        
        printf("\r\n\r\n");
    }
   
    while (1); // end the program safely by infinit loop
    return 0;
}